function x = robust_SL0(A, y, sigmaNoise, varargin)
% Modified SL0 algorithm. Should be robust to noise.
%
% x = robust_SL0(A, y, sigmaNoise, varargin)
%
% Input: A - 'mixing' matrix
%
%        y - right side vector (y = Ax + noise)
%
%        sigmaNoise - approximate noise level
%
%        optional parameters: ...
%
% Input optional parameters as ('parameter name', 'parameter value') pairs
%
% Output: x - sparse solution
%
% [1] Eftekhari, Babaie-Zadeh, Jutten, Moghaddam, Robust-SL0 for stable
% sparse representations in noisy settings, ICASSP 2009.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
inner_iters = 5;
sigmaMin = 1e-6;
decrease = 0.8;
mi = 0.1;
projection = 'default';
verbose = false;

if isfloat(A)
    D = @(x) A*x;       
    Apinv = pinv(A);
    Dpinv = @(x) Apinv*x;
elseif isa(A, 'function_handle') 
    D = A;    
else
    error('A is of wrong type!');
end

% Check for optional parameters
if(mod(length(varargin),2) == 1)
    error('Optional arguments come in pairs.');
end
for i = 1:2:length(varargin)
    switch lower(varargin{i})

        case 'apinv'
            Dpinv = varargin{i+1};
        case 'mi'
            mi = varargin{i+1};
        case 'inner'
            inner_iters = varargin{i+1};
        case 'decrease'
            decrease = varargin{i+1};
        case 'sigmamin'
            sigmaMin = varargin{i+1};
        case 'projection'
            projection = varargin{i+1};
        case 'verbose'
            verbose = varargin{i+1};
            
    end
end

% Check if Apinv is provided
if ~exist('Dpinv', 'var')
    error('Pseudoinverse operator not provided!');
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
x = Dpinv(y);
sigma = 2*max(abs(x));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main algorithm
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
iter = 0;
if verbose
    hh = waitbar(0, 'Algorithm in progress...');
end
maxiter = ceil(log(sigmaMin/sigma) / log(decrease));

while sigma > sigmaMin
    iter = iter+1;
    
    for i = 1:inner_iters
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Gradient descent step
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%        
        delta = OurDelta(x, sigma);
        x = x - mi * delta;
        
        residual = D(x) - y;                     
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Projection step
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % Project if norm of the residual is above the noise level (this is
        % the approach suggested in [1])
        
        if norm(residual, 2)^2 > sigmaNoise
              x = x - Dpinv(residual);
        end        
        
            
        
    end    
    
    sigma = sigma * decrease;
    
    if verbose
        waitbar(iter/maxiter, hh);
    end
    
end

if verbose
    close(hh)
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function delta = OurDelta(s, sigma)

% delta = s.*exp(-s.^2/sigma^2);
delta = s .* exp(-s.^2/(2*sigma^2));

